
/****************************************************************************
* FILE : pas2ps.rd
* NAME: Dulith Herath.
* INPUT/OUTPUT  : rd++ input file (partial)
* PROJECT: CS 541 CS-University of Kentucky
* SOURCE:   Provided as a handout by Jerzy W. Jaromczyk
*           Additionals and modifications by Dulith Herath
* COPYRIGHT: University of Kentucky @1999


* ADD:  see file README for requirements
* LL(1) grammar for simple Pascal-like language based on Aho-Sethi-Ullman
* NOTE: some code (semantic actions) was removed, without removing the relevant
*       declarations, so there are a few seemingly useless declarations.
****************************************************************************/

%{

#include <stdio.h>
#include <iostream.h>
#include <malloc.h>
#include "st_stack.h"
#include "sym_table.h"

/* lex associated functions */
extern char yytext[];
extern int  yyleng;
extern "C" int yylex(void);

extern char *id_lookback;

/* output stream. */
extern FILE *out_stream;

/* flags for output options during parsing. */
extern int debug_flag;

/* The lookahead */
extern int lookahead;

/* Compliance. */
int compile_errors = 0;
int current_offset  = 0;

extern char percent;

char *main_program_name;

/* The symbol table stack. */
ST_Stack st_stack;

/* The symbol tables */
Symbol_Table *GlobalSymbolTable;
Symbol_Table *CurrentSymbolTable = NULL;
Symbol_Table *TempSymbolTable;
/* Current Symbol */
LL_Node *CurrentSymbol = NULL;
LL_Node *CurrentProc = NULL;

S_Stack function_stack;
S_Stack param_stack;
S_Stack variable_stack;
S_Stack temp_stack;
S_Stack procedure_call_stack;
S_Stack global_stack;
bool index;
bool declaration_flag = false;
bool assign_flag = false;
bool global_flag = true;
char temp_proc_name[128];
char last_fcn[128];

#define DEBUG   0
#define DPRINT(call)   { if(DEBUG) { call ; } }

%}

%start Program
%token PROGRAM PROCEDURE FUNCTION
%token BEGIN_BLOCK END_BLOCK
%token WHILE DO IF THEN ELSE ARRAY OF VAR REAL INTEGER
%token <Op>RELOP MULOP ADDOP SIGN
%token MINUS PLUS MULTIPLY DIV DIVIDE MOD AND NOT OR
%token <Integer>INT_NUM
%token REAL_NUM
%token ID
%token EQ GE GT LT LE NE
%token COLON COMMA SEMICOLON ASSIGN LBRACK RBRACK LPAREN RPAREN RANGE PERIOD
%token READ WRITE WRITELN
%token UNKNOWN
%token FOR TO

%union {
   int          Op;
   double       Float;
   int          Integer;
};

%%

Program:
#   GlobalSymbolTable = new Symbol_Table;
#   TempSymbolTable = new Symbol_Table;
#   /* Output a generic PS prolog. */
#   fprintf(out_stream, "/fontsize 12 def\n");
#   fprintf(out_stream, "/default_font { /Helvetica findfont fontsize scalefont setfont } def\n");
#   fprintf(out_stream, "/LM 12 def  %c left margin\n",percent);
#   fprintf(out_stream, "/TM 750 def %c top margin\n",percent);
#   fprintf(out_stream, "/CP TM def %c current position\n",percent);
#   fprintf(out_stream, "/reset { showpage default_font LM TM moveto } def\n");
#   fprintf(out_stream, "/write_line_string 255 string def\n");
#   fprintf(out_stream, "default_font LM TM moveto\n");
/*
#   fprintf(out_stream, "/stackit { } def \n");
#   CurrentSymbol = GlobalSymbolTable->Insert("stackit");
#   CurrentSymbol->SetDefinitionType(T_PROCEDURE);
*/
#   CurrentSymbol = GlobalSymbolTable->Insert("reset");
#   CurrentSymbol->SetDefinitionType(T_PROCEDURE);
#   fprintf(out_stream, "%c Begin Program \n",percent);
#   /* Set up the current symbol table to be the global symbol table. */
#   CurrentSymbolTable = GlobalSymbolTable;
#   st_stack.Push(GlobalSymbolTable);
#   DPRINT(fprintf(stderr, "Program: PROGRAM ID SEMICOLON Block PERIOD\n"));
    PROGRAM ID SEMICOLON Block PERIOD
#   GlobalSymbolTable->FPrintF(stderr);
;

Block:
#               DPRINT(fprintf(stderr, "Block: ALLDeclaration CompoundStatement\n"));
                AllDeclaration CompoundStatement
;

AllDeclaration:
#             DPRINT(fprintf(stderr, "AllDeclaration: Declarations SubprogramDeclarations\n"));
#             variable_stack.Push(NULL);
              Declarations
#             /* Now we know the dict size so we can set up the new dict. */
#             if(CurrentSymbolTable != GlobalSymbolTable) {
#                  fprintf(out_stream, "%c Set new frame with %i slots\n", percent, current_offset);
#                  fprintf(out_stream, " %c and intitialize variables to 0\n", percent, current_offset);
#                  fprintf(out_stream, " %i dict begin\n", current_offset);
#             }
#             /* Initialize the VAR delclarations. */
#             while((CurrentSymbol = variable_stack.Pop()) != NULL) {
#                  if(CurrentSymbol->GetDefinitionType() == T_VAR) {
#                    fprintf(out_stream, " /%s 0 def \n",CurrentSymbol->GetName());
#		   } else {
#                    fprintf(out_stream, " /%s %i array def  \n",CurrentSymbol->GetName(),CurrentSymbol->GetDimension());
#		   }
#                  if(CurrentSymbolTable == GlobalSymbolTable) {
#                    global_stack.Push(CurrentSymbol);
#                  }
#             }
	      SubprogramDeclarations
;

Declarations:
#               DPRINT(fprintf(stderr, "Declarations: VAR IdentifierList COLON Type SEMICOLON Declarations \n"));
                VAR
#	        declaration_flag = true;
		IdentifierList COLON Type SEMICOLON Declarations
                |
#               DPRINT(fprintf(stderr, "Declarations: RD_EMPTY\n"));
                RD_EMPTY
;

IdentifierList:
#               current_offset++;
#               CurrentSymbol = CurrentSymbolTable->Insert(id_lookback);
#               CurrentSymbol->SetOffset(current_offset);
#               CurrentSymbol->SetBlockLevel(st_stack.GetSize());
#               temp_stack.Push(CurrentSymbol);
#               DPRINT(fprintf(stderr, "IdentifierList: ID IdListTail\n"));
                ID IdListTail
;

IdListTail:
#               DPRINT(fprintf(stderr, "IdListTail: COMMA ID IdListTail\n"));
                COMMA
#               current_offset++;
#               CurrentSymbol = CurrentSymbolTable->Insert(id_lookback);
#               CurrentSymbol->SetOffset(current_offset);
#               CurrentSymbol->SetBlockLevel(st_stack.GetSize());
#               temp_stack.Push(CurrentSymbol);
		ID IdListTail
	        |
#               DPRINT(fprintf(stderr, "IdListTail: RD_EMPTY\n"));
                RD_EMPTY
;

Type:
#               DPRINT(fprintf(stderr, "Type: StandardType\n"));
		StandardType
#		/* Set the type. */
#               while((CurrentSymbol = temp_stack.Pop()) != NULL) {
#                      CurrentSymbol->SetDefinitionType(T_VAR);
#                      variable_stack.Push(CurrentSymbol);
#               }
	        |
#               DPRINT(fprintf(stderr, "Type: ARRAY LBRACK INT_NUM RANGE INT_NUM RBRACK OF StandardType\n"));
  	        ARRAY LBRACK INT_NUM RANGE INT_NUM RBRACK OF StandardType
#               /* Set type and dimension. */
#               while((CurrentSymbol = temp_stack.Pop()) != NULL) {
#		       CurrentSymbol->SetDefinitionType(T_ARRAY);
#                      CurrentSymbol->SetDimension($5);
#                      variable_stack.Push(CurrentSymbol);
#               }
;

StandardType:
#               DPRINT(fprintf(stderr, "StandardType: INTEGER\n"));
                INTEGER
	        |
#               DPRINT(fprintf(stderr, "StandardType: REAL\n"));
             	REAL
;

SubprogramDeclarations:
#               DPRINT(fprintf(stderr, "SubprogramDeclarations: SubprogramDeclaration SEMICOLON SubprogramDeclarations\n"));
                SubprogramDeclaration SEMICOLON
#               /* Insure global variables set, Ensure if over 65.*/
#               LL_Node *temp_node;
#               global_stack.First();
#               while(global_stack.GetCurrent() != NULL) {
#                 temp_node = global_stack.GetCurrent()->GetSymbol();
#                 fprintf(out_stream, " %s ", temp_node->GetName());
#                 temp_stack.Push(temp_node);
#                 global_stack.Next();
#               }
#               fprintf(out_stream, " end\n");
#               global_stack.First();
#               while((temp_node = temp_stack.Pop()) != NULL) {
#                 fprintf(out_stream, " /%s exch def\n", temp_node->GetName());
#               }
#               fprintf(out_stream, "  } store\n");
		SubprogramDeclarations
                |
#               DPRINT(fprintf(stderr, "SubprogramDeclarations: RD_EMPTY\n"));
              	RD_EMPTY
;

SubprogramDeclaration:
#               DPRINT(fprintf(stderr, "SubprogramDeclaration: SubprogramHead AllDeclaration CompoundStatement\n"));
                SubprogramHead AllDeclaration
#               /* No local variables, so we didn't go into AllDeclaration. */
#               /* We need to setup up the dictionary for this proc. */
#               if(declaration_flag == false) {
#                  fprintf(out_stream, " %i dict begin\n", current_offset);
#               }
#               /* Do we have parameters??? */
#               while((CurrentSymbol = variable_stack.Pop()) != NULL) {
#                  if(CurrentSymbol->GetDefinitionType() == T_VAR) {
#		     fprintf(out_stream," /%s exch def \n",CurrentSymbol->GetName());
#		   } else {
#                    fprintf(out_stream," %i array astore /%s exch def \n",CurrentSymbol->GetDimension(),CurrentSymbol->GetName());
#		   }
#               }
		CompoundStatement
#	        /* Pop the current symbol table. */
#               CurrentSymbolTable->FPrintF(stderr);
#	        CurrentSymbolTable = st_stack.Pop();
;

SubprogramHead:
#               DPRINT(fprintf(stderr, "SubprogramHead: FUNCTION ID Arguments COLON StandardType SEMICOLON\n"));
                FUNCTION ID
#               declaration_flag = false;
#	        current_offset = 0;
#               CurrentProc = CurrentSymbolTable->Insert(id_lookback);
#		/* Get a new symbol table and push the current onto the stack. */
#		CurrentProc->SetDefinitionType(T_FUNCTION);
#               strcpy (last_fcn,"");
#               strcpy (last_fcn,CurrentProc->GetName());
#               CurrentProc->SetBlockLevel(st_stack.GetSize());
#               fprintf(out_stream, " /%s { ", id_lookback);
#               /* Put a new symbol table on the symbol table stack. */
#               CurrentSymbolTable = new Symbol_Table;
#               st_stack.Push(CurrentSymbolTable);
		Arguments
#               CurrentProc->SetSize(current_offset);
		COLON StandardType SEMICOLON
	        |

#               DPRINT(fprintf(stderr, "SubprogramHead: PROCEDURE ID Arguments SEMICOLON\n"));
             	PROCEDURE ID
#               declaration_flag = false;
#	        current_offset = 0;
#               CurrentProc = CurrentSymbol = CurrentSymbolTable->Insert(id_lookback);
#		/* Get a new symbol table and push the current onto the stack. */
#		CurrentProc->SetDefinitionType(T_PROCEDURE);
#               CurrentProc->SetBlockLevel(st_stack.GetSize());
#               fprintf(out_stream, " /%s { ", id_lookback);
#               /* Put a new symbol table on the symbol table stack. */
#               CurrentSymbolTable = new Symbol_Table;
#               st_stack.Push(CurrentSymbolTable);
		Arguments
#               CurrentProc->SetSize(current_offset);
		SEMICOLON
;

Arguments:
#             DPRINT(fprintf(stderr, "Arguments: LPAREN ParameterList RPAREN\n"));
#             if(variable_stack.GetSize() > 0) {
#                variable_stack.Push(NULL);
#               }
              LPAREN ParameterList RPAREN
	      |
#             DPRINT(fprintf(stderr, "Arguments: RD_EMPTY\n"));
              RD_EMPTY
;

ParameterList:
#             DPRINT(fprintf(stderr, "ParameterList: IdentifierList COLON Type ParameterTail\n"));
              IdentifierList COLON Type ParameterTail
;

ParameterTail:
#             DPRINT(fprintf(stderr, "ParameterTail: SEMICOLON IdentifierList COLON Type ParameterTail\n"));
              SEMICOLON IdentifierList COLON Type ParameterTail
              |
#             DPRINT(fprintf(stderr, "ParameterTail: RD_EMPTY\n"));
	      RD_EMPTY
;

CompoundStatement:
#             DPRINT(fprintf(stderr, "CompoundStatement: BEGIN_BLOCK OptionalStatement END_BLOCK\n"));
              BEGIN_BLOCK OptionalStatement
#             strcpy (last_fcn,"empty_");
              END_BLOCK
;

OptionalStatement:
#             DPRINT(fprintf(stderr, "OptionalStatement: StatementList\n"));
              StatementList
	      |
#             DPRINT(fprintf(stderr, "OptionalStatement: RD_EMPTY\n"));
              RD_EMPTY
;

StatementList:
#             DPRINT(fprintf(stderr, "StatementList: RD_EMPTY\n"));
              Statement StatementListTail
;

StatementListTail:
#             DPRINT(fprintf(stderr, "StatementListTail: SEMICOLON Statement StatementListTail\n"));
              SEMICOLON Statement StatementListTail
	      |
#             DPRINT(fprintf(stderr, "StatementListTail: RD_EMPTY\n"));
       	      RD_EMPTY
;

Statement:
#             DPRINT(fprintf(stderr, "Statement: ID StatementOp\n"));
	      ID
#             CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
#             if((CurrentSymbol == NULL)&&(CurrentSymbolTable != GlobalSymbolTable))
#                       CurrentSymbol = GlobalSymbolTable->Lookup(id_lookback, -1);
	      StatementOp
	      |
#             DPRINT(fprintf(stderr, "Statement: IfStatement\n"));
  	      IfStatement
	      |
#             DPRINT(fprintf(stderr, "Statement: CompoundStatement\n"));
  	      CompoundStatement
	      |
#             DPRINT(fprintf(stderr, "Statement: WhileStatement\n"));
   	      WhileStatement
	      |
#	      DPRINT(fprintf(stderr, "Statement: ForStatement\n"));
	      ForStatement
	      |
#             DPRINT(fprintf(stderr, "Statement: ReadStatement\n"));
   	      ReadStatement
	      |
#             DPRINT(fprintf(stderr, "Statement: WriteStatement\n"));
   	      WriteStatement
	      |
#             DPRINT(fprintf(stderr, "Statement: WriteLnStatement\n"));
   	      WriteLnStatement
;

StatementOp:
#             DPRINT(fprintf(stderr, "StatementOp: Index OptionalAssign\n"));
#             if(CurrentSymbol == NULL) {
#                fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
#                compile_errors++;
#             } else {
#                fprintf(out_stream, " %s ", CurrentSymbol->GetName());
#             }
              Index OptionalAssign
#             fprintf(out_stream, " put\n");
	      |
#             DPRINT(fprintf(stderr, "StatementOp: ProcTail OptionalAssign\n"));
#             // :o) storing a procedure name which is not defined
#             //procedure_call_stack.Push(CurrentSymbol);
#             CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
#             if (CurrentSymbol!=NULL) strcpy (id_lookback,"");
#             procedure_call_stack.Push(CurrentSymbol);
#             strcpy (temp_proc_name,"");
#             strcpy (temp_proc_name,id_lookback);
	      ProcTail
#             if (CurrentSymbol==NULL)
#                {
#                if (strcmp(temp_proc_name,last_fcn)!=0)
#                fprintf (out_stream, "%s \n", temp_proc_name);
#                }
              OptionalAssign

	      |
#             DPRINT(fprintf(stderr, "StatementOp: AssignStatement\n"));
#             assign_flag = false;
#             if(CurrentSymbol == NULL) {
#                fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
#                compile_errors++;
#             } else {
#                /* Check for function return values. */
#                if(CurrentSymbol->GetDefinitionType() == T_PROCEDURE) {
#                   fprintf(stderr, "%i: Attempt to return from procedure: %s\n",rdlines,CurrentSymbol->GetSpelling());
#                   compile_errors++;
#                }
#                if(CurrentSymbol->GetDefinitionType() != T_FUNCTION) {
#                    fprintf(out_stream, " /%s ", CurrentSymbol->GetName());
#                    assign_flag = true;
#                }
#             }
	      AssignStatement
#             if(assign_flag == true) {
#                fprintf(out_stream, " store\n");
#             }
	      |
#             DPRINT(fprintf(stderr, "StatementOp: RD_EMPTY\n"));
	      RD_EMPTY
;

OptionalAssign:
#             DPRINT(fprintf(stderr, "OptionalAssign: AssignStatement\n"));
#             /* The RHS is the return value from this function. */
	      AssignStatement
	      |
#             DPRINT(fprintf(stderr, "OptionalAssign: RD_EMPTY\n"));
#             CurrentSymbol = procedure_call_stack.Pop();
#             if(CurrentSymbol != NULL) {
#                fprintf(out_stream, " %s ",CurrentSymbol->GetSpelling());
#             }
	      RD_EMPTY
;

AssignStatement:
#             DPRINT(fprintf(stderr, "AssignStatement: ASSIGN Expression\n"));
              ASSIGN Expression
;

WriteStatement:
#             DPRINT(fprintf(stderr, "WriteStatement: WRITE LPAREN Expression RPAREN\n"));
              WRITE LPAREN Expression RPAREN
#             /* Ok we want to write the top of the stack. */
#             fprintf(out_stream, "write_line_string cvs ");
#             fprintf(out_stream, " show\n");
;


WriteLnStatement:
#             DPRINT(fprintf(stderr, "WriteStatement: WRITELN LPAREN Expression RPAREN\n"));
              WRITELN LPAREN Expression
		RPAREN
#             /* Ok we want to write the top of the stack. */
#             fprintf(out_stream, "write_line_string cvs ");
#             fprintf(out_stream, " show\n");
#             fprintf(out_stream, " /CP CP fontsize sub 1 sub def\n");
#             fprintf(out_stream, " LM CP moveto\n");
;

ReadStatement:
#             DPRINT(fprintf(stderr, "ReadStatement: READ LPAREN IdRList RPAREN\n"));
              READ LPAREN IdRList RPAREN
;

IdRList:
#             DPRINT(fprintf(stderr, "IdRList: IdAccess IdRTail\n"));
              IdAccess IdRTail
;

IdRTail:
#             DPRINT(fprintf(stderr, "IdRTail: COMMA IdAccess IdRTail\n"));
              COMMA IdAccess IdRTail
	      |
#             DPRINT(fprintf(stderr, "IdRTail: RD_EMPTY\n"));
              RD_EMPTY
;

Index:
#             DPRINT(fprintf(stderr, "Index: LBRACK IndexType RBRACK\n"));
              LBRACK IndexType RBRACK
;

IndexType:
#             DPRINT(fprintf(stderr, "IndexType: INT_NUM\n"));
	      INT_NUM
#             fprintf(out_stream, " %i 1 sub\n", $1);
	      |
#             DPRINT(fprintf(stderr, "IndexType: ID\n"));
	      ID
#             CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
#             if((CurrentSymbol == NULL)&&(CurrentSymbolTable != GlobalSymbolTable))
#                       CurrentSymbol = GlobalSymbolTable->Lookup(id_lookback, -1);
#             if(CurrentSymbol == NULL) {
#                fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
#                compile_errors++;
#             } else {
#                fprintf(out_stream, " %s 1 sub\n", CurrentSymbol->GetName());
#             }
;

IdAccess:
#             DPRINT(fprintf(stderr, "IdAccess: ID IdAccessTail\n"));
              ID IdAccessTail
;

IdAccessTail:
#             DPRINT(fprintf(stderr, "IdAccessTail: Index\n"));
              Index
	      |
#             DPRINT(fprintf(stderr, "IdAccessTail: RD_EMPTY\n"));
 	      RD_EMPTY
;

WhileStatement:
#	      fprintf(out_stream, "{ %c beginning of while loop\n", percent);
#             DPRINT(fprintf(stderr, "WhileStatement: WHILE Expression DO \n"));
 	      WHILE Expression DO
#	      fprintf(out_stream, "\n  { ");
#             DPRINT(fprintf(stderr, "WhileStatement: Statement\n"));
	      Statement
#	      fprintf(out_stream, "  }\n {exit} ifelse\n} loop %c  end of while\n", percent);
;

ForStatement:
	      FOR ID AssignStatement
	      TO
#	      fprintf (out_stream, "1 "); //assuming step is 1
	      Expression
	      DO
#             fprintf (out_stream, "{");
	      Statement
#	      fprintf (out_stream, "} for \n");
;

IfStatement:
#             DPRINT(fprintf(stderr, "IfStatement: IF Expression THEN \n"));
              IF Expression THEN
#             fprintf (out_stream,"{");
	      Statement
#             fprintf (out_stream,"}\n");
	      Testelse
;

Testelse:
#	      DPRINT(fprintf(stderr, "else or no else \n"));
              ELSE
#	      fprintf (out_stream,"{");
              Statement
#	      fprintf (out_stream,"}\n");
#	      fprintf (out_stream," ifelse ");
	      |
	      SEMICOLON

;

ProcTail:
#             DPRINT(fprintf(stderr, "ProcTail: LPAREN ProcExpressionList RPAREN\n"));
              LPAREN ProcExpressionList RPAREN
;

ProcExpressionList:
#             DPRINT(fprintf(stderr, "ProcExpressionList: Expression ProcExpressionListTail\n"));
              Expression ProcExpressionListTail
#             while((CurrentSymbol = function_stack.Pop()) != NULL) {
#                    fprintf(out_stream, " %s ",CurrentSymbol->GetName());
#                   }
#             while((CurrentSymbol = procedure_call_stack.Pop()) != NULL){
#                    fprintf(out_stream, " %s ",CurrentSymbol->GetName());
#                   }
;

ProcExpressionListTail:
#             DPRINT(fprintf(stderr, "ProcExpressionListTail: COMMA Expression ProcExpressionListTail\n"));
              COMMA Expression ProcExpressionListTail
	      |
#             DPRINT(fprintf(stderr, "ProcExpressionListTail: RD_EMPTY\n"));
              RD_EMPTY
;

Expression:
#             DPRINT(fprintf(stderr, "Expression: SimpleExpression ExpressionTail\n"));
              SimpleExpression ExpressionTail
              |
#             DPRINT(fprintf(stderr, "Expression: RD_EMPTY\n"));
   	      RD_EMPTY

;

ExpressionTail:
#             DPRINT(fprintf(stderr, "ExpressionTail: RELOP SimpleExpression\n"));
              RELOP SimpleExpression
#             switch($1) {
#                case EQ:
#                      fprintf(out_stream, " eq ");
#                      break;
#                case GE:
#                      fprintf(out_stream, " ge ");
#                      break;
#                case GT:
#                      fprintf(out_stream, " gt ");
#                      break;
#                case LT:
#                      fprintf(out_stream, " lt ");
#                      break;
#                case LE:
#                      fprintf(out_stream, " le ");
#                      break;
#                case NE:
#                      fprintf(out_stream, " ne ");
#                      break;
#	      }
              |
#             DPRINT(fprintf(stderr, "ExpressionTail: RD_EMPTY\n"));
              RD_EMPTY
;

SimpleExpression:
#             DPRINT(fprintf(stderr, "SimpleExpression: SIGN Term SimpleExpressionTail\n"));
              SIGN Term SimpleExpressionTail
#             if($1 == MINUS) {
#                fprintf(out_stream, " neg\n");
#             }
	      |
#             DPRINT(fprintf(stderr, "SimpleExpression: Term SimpleExpressionTail\n"));
	      Term
               SimpleExpressionTail
;

SimpleExpressionTail:
#             DPRINT(fprintf(stderr, "SimpleExpressionTail: ADDOP Term SimpleExpressionTail\n"));
              ADDOP Term SimpleExpressionTail
#             switch($1) {
#                case PLUS:
#                      fprintf(out_stream, " add\n");
#                      break;
#                case MINUS:
#                      fprintf(out_stream, " sub\n");
#                      break;
#                case OR:
#                      fprintf(out_stream, " or\n");
#                      break;
#             }
	      |
#             DPRINT(fprintf(stderr, "SimpleExpressionTail: RD_EMPTY\n"));
 	      RD_EMPTY
;

Term:
#             DPRINT(fprintf(stderr, "Term: Factor TermTail\n"));
              Factor TermTail
;


Sign:
#     DPRINT(fprintf(stderr, "Sign: SIGN\n"));
      SIGN
#       if($1 == MINUS) {
#           fprintf(out_stream, " neg\n");
#       }
      |
#     DPRINT(fprintf(stderr, "Sign: RD_EMPTY\n"));
      RD_EMPTY
;

TermTail:
#             DPRINT(fprintf(stderr, "TermTail: MULOP Sign Factor TermTail\n"));
              MULOP Sign Factor TermTail
#             switch($1) {
#                case MULTIPLY:
#                      fprintf(out_stream, " mul\n");
#                      break;
#                case DIVIDE:
#                      fprintf(out_stream, " div\n");
#                      break;
#                case MOD:
#                      fprintf(out_stream, " mod\n");
#                      break;
#                case DIV:
#                      fprintf(out_stream, " idiv\n");
#                      break;
#                case AND:
#                      fprintf(out_stream, " and\n");
#                      break;
#             }
	      |
#             DPRINT(fprintf(stderr, "TermTail: RD_EMPTY\n"));
 	      RD_EMPTY
;

Factor:
#             DPRINT(fprintf(stderr, "Factor: ID FactorOp\n"));
#             index = false;
#             CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
#             if((CurrentSymbol == NULL)&&(CurrentSymbolTable != GlobalSymbolTable))
#                       CurrentSymbol = GlobalSymbolTable->Lookup(id_lookback, -1);
#             if(CurrentSymbol == NULL) {
#                fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
#                compile_errors++;
#             } else {
#                if(CurrentSymbol->GetDefinitionType() != T_FUNCTION && CurrentSymbol->GetDefinitionType() != T_PROCEDURE) {
#                    fprintf(out_stream, " %s ",CurrentSymbol->GetName());
#                    }
#  // pushing function names into a function stack
#                if (CurrentSymbol->GetDefinitionType() ==T_FUNCTION)
#		     {
#		     function_stack.Push(CurrentSymbol);
#		     }
#  // pushing procedure names into a procedure stack
#                if (CurrentSymbol->GetDefinitionType() ==T_PROCEDURE)
#                    {
#                    procedure_call_stack.Push(CurrentSymbol);
#                    }
#             }
	      ID FactorOp
#             if(index == true) {
#               /* I think the value of index should be on the PS stack. */
#		fprintf(out_stream, " get ");
#             }
	      |
#             DPRINT(fprintf(stderr, "Factor: NOT Factor\n"));
              NOT Factor
#             fprintf(out_stream, " not\n");
	      |
#             DPRINT(fprintf(stderr, "Factor: INT_NUM\n"));
   	      INT_NUM
#             fprintf(out_stream, " %i ",$1);
	      |
#             DPRINT(fprintf(stderr, "Factor: REAL_NUM\n"));
  	      REAL_NUM
#             fprintf(out_stream, " %g ",$1);
	      |
#             DPRINT(fprintf(stderr, "Factor: LPAREN Expression RPAREN\n"));
  	      LPAREN Expression RPAREN
;

FactorOp:
#             DPRINT(fprintf(stderr, "FactorOp: Index\n"));
#             index = true;
  	      Index
	      |
#             DPRINT(fprintf(stderr, "FactorOp: ProcTail\n"));
   	      ProcTail
	      |
#             DPRINT(fprintf(stderr, "FactorOp: RD_EMPTY\n"));
	      RD_EMPTY
;
