#include "rdparser.h"

#include <stdio.h>
#include <iostream>
#include <malloc.h>
#include "st_stack.h"
#include "sym_table.h"

extern char *id_lookback;

/* flags for output options during parsing. */
extern int debug_flag;
/* The lookahead */
extern int lookahead;
/* Compliance. */
int compile_errors = 0;
int current_offset  = 0;

char *main_program_name;
/* The symbol table stack. */
ST_Stack st_stack;
/* The symbol tables */
Symbol_Table *GlobalSymbolTable;
Symbol_Table *CurrentSymbolTable = NULL;
Symbol_Table *TempSymbolTable;
/* Current Symbol */
LL_Node *CurrentSymbol = NULL;
LL_Node *CurrentProc = NULL;
S_Stack function_stack;
S_Stack param_stack;
S_Stack variable_stack;
S_Stack temp_stack;
S_Stack procedure_call_stack;
S_Stack global_stack;
bool rdindex;
bool declaration_flag = false;
bool assign_flag = false;
bool global_flag = true;
char temp_proc_name[128];
char last_fcn[128];
#define DEBUG   0
#define DPRINT(call)   { if(DEBUG) { call ; } }
rdunion rdlval, rdoldval;

int rdlines=1;
RdParser::RdParser() {
  rdtoken=new RdToken;
}

RdParser::~RdParser() {
  delete rdtoken;
}

rdunion RdParser::rdparse() {
  rdunion oldrdlval, oldrdoldval;
  rdunion ret, s1;
  int oldrdlines;
  oldrdlval=rdlval;
  oldrdoldval=rdoldval;
  oldrdlines=rdlines;
  ret=parseProgram(&s1);
  rdlval=oldrdlval;
  rdoldval=oldrdoldval;
  rdlines=oldrdlines;
  return ret;
}

rdunion RdParser::parseProgram(rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5;

  rdtoken->getFirst(&RdToken::inFirstProgram,&RdToken::inFollowProgram,&RdToken::inAugmentProgram);
  if(rdtoken->inFirst_PROGRAM_ID_SEMICOLON_Block_PERIOD()) {
    GlobalSymbolTable = new Symbol_Table;
    TempSymbolTable = new Symbol_Table;
    /* Output a generic PS prolog. */
    printf("/fontsize 12 def\n");
    printf("/default_font { /Helvetica findfont fontsize scalefont setfont } def\n");
    printf("/LM 12 def  %% left margin\n");
    printf("/TM 750 def %% top margin\n");
    printf("/CP TM def %% current position\n");
    printf("/reset { showpage default_font LM TM moveto } def\n");
    printf("/write_line_string 255 string def\n");
    printf("default_font LM TM moveto\n");
    CurrentSymbol = GlobalSymbolTable->Insert("reset");
    CurrentSymbol->SetDefinitionType(T_PROCEDURE);
    printf("%% Begin Program \n");
    /* Set up the current symbol table to be the global symbol table. */
    CurrentSymbolTable = GlobalSymbolTable;
    st_stack.Push(GlobalSymbolTable);
    DPRINT(fprintf(stderr, "Program: PROGRAM ID SEMICOLON Block PERIOD\n"));
    rdtoken->advance(PROGRAM);
    synth1=rdoldval;
    rdtoken->advance(ID);
    synth2=rdoldval;
    rdtoken->advance(SEMICOLON);
    synth3=rdoldval;
    synth4=parseBlock(&rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
#if 1
    rdtoken->advance(PERIOD);
    synth5=rdoldval;
#endif
    GlobalSymbolTable->FPrintF(stderr);
  }
#if 1
  rdtoken->getFollow(&RdToken::inFollowProgram,&RdToken::inAugmentProgram);
#endif
  return *retval;
}

rdunion RdParser::parseBlock(rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstBlock,&RdToken::inFollowBlock,&RdToken::inAugmentBlock);
  if(rdtoken->inFirst_AllDeclaration_CompoundStatement()) {
    DPRINT(fprintf(stderr, "Block: ALLDeclaration CompoundStatement\n"));
    synth1=parseAllDeclaration(&rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseCompoundStatement(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowBlock,&RdToken::inAugmentBlock);
  return *retval;
}

rdunion RdParser::parseAllDeclaration(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstAllDeclaration,&RdToken::inFollowAllDeclaration,&RdToken::inAugmentAllDeclaration);
  if(rdtoken->inFirst_Declarations_SubprogramDeclarations()) {
    DPRINT(fprintf(stderr, "AllDeclaration: Declarations SubprogramDeclarations\n"));
    variable_stack.Push(NULL);
    synth1=parseDeclarations(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
    /* Now we know the dict size so we can set up the new dict. */
    if(CurrentSymbolTable != GlobalSymbolTable) {
      printf("%% Set new frame with %i slots\n", current_offset);
      printf(" %% and intitialize variables to 0\n");
      printf(" %i dict begin\n", current_offset);
    }
    /* Initialize the VAR delclarations. */
    while((CurrentSymbol = variable_stack.Pop()) != NULL) {
      if(CurrentSymbol->GetDefinitionType() == T_VAR) {
	printf(" /%s 0 def \n",CurrentSymbol->GetName());
      } else {
	printf(" /%s %i array def  \n",CurrentSymbol->GetName(),CurrentSymbol->GetDimension());
      }
      if(CurrentSymbolTable == GlobalSymbolTable) {
	global_stack.Push(CurrentSymbol);
      }
    }
    synth2=parseSubprogramDeclarations(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowAllDeclaration,&RdToken::inAugmentAllDeclaration);
  return *retval;
}

rdunion RdParser::parseDeclarations(rdunion *inherit6, rdunion *inherit5, rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5, synth6;

  rdtoken->getFirst(&RdToken::inFirstDeclarations,&RdToken::inFollowDeclarations,&RdToken::inAugmentDeclarations);
  if(rdtoken->inFirst_VAR_IdentifierList_COLON_Type_SEMICOLON_Declarations()) {
    DPRINT(fprintf(stderr, "Declarations: VAR IdentifierList COLON Type SEMICOLON Declarations \n"));
    rdtoken->advance(VAR);
    synth1=rdoldval;
    declaration_flag = true;
    synth2=parseIdentifierList(&rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(COLON);
    synth3=rdoldval;
    synth4=parseType(&rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
    rdtoken->advance(SEMICOLON);
    synth5=rdoldval;
    synth6=parseDeclarations(&rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4, &synth5,&synth6);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "Declarations: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowDeclarations,&RdToken::inAugmentDeclarations);
  return *retval;
}

rdunion RdParser::parseIdentifierList(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstIdentifierList,&RdToken::inFollowIdentifierList,&RdToken::inAugmentIdentifierList);
  if(rdtoken->inFirst_ID_IdListTail()) {
    current_offset++;
    CurrentSymbol = CurrentSymbolTable->Insert(id_lookback);
    CurrentSymbol->SetOffset(current_offset);
    CurrentSymbol->SetBlockLevel(st_stack.GetSize());
    temp_stack.Push(CurrentSymbol);
    DPRINT(fprintf(stderr, "IdentifierList: ID IdListTail\n"));
    rdtoken->advance(ID);
    synth1=rdoldval;
    synth2=parseIdListTail(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowIdentifierList,&RdToken::inAugmentIdentifierList);
  return *retval;
}

rdunion RdParser::parseIdListTail(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstIdListTail,&RdToken::inFollowIdListTail,&RdToken::inAugmentIdListTail);
  if(rdtoken->inFirst_COMMA_ID_IdListTail()) {
    DPRINT(fprintf(stderr, "IdListTail: COMMA ID IdListTail\n"));
    rdtoken->advance(COMMA);
    synth1=rdoldval;
    current_offset++;
    CurrentSymbol = CurrentSymbolTable->Insert(id_lookback);
    CurrentSymbol->SetOffset(current_offset);
    CurrentSymbol->SetBlockLevel(st_stack.GetSize());
    temp_stack.Push(CurrentSymbol);
    rdtoken->advance(ID);
    synth2=rdoldval;
    synth3=parseIdListTail(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "IdListTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowIdListTail,&RdToken::inAugmentIdListTail);
  return *retval;
}

rdunion RdParser::parseType(rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5, synth6, synth7, synth8;

  rdtoken->getFirst(&RdToken::inFirstType,&RdToken::inFollowType,&RdToken::inAugmentType);
  if(rdtoken->inFirst_StandardType()) {
    DPRINT(fprintf(stderr, "Type: StandardType\n"));
    synth1=parseStandardType(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
    /* Set the type. */
    while((CurrentSymbol = temp_stack.Pop()) != NULL) {
      CurrentSymbol->SetDefinitionType(T_VAR);
      variable_stack.Push(CurrentSymbol);
    }
  } else if(rdtoken->inFirst_ARRAY_LBRACK_INT95NUM_RANGE_INT95NUM_RBRACK_OF_StandardType()) {
    DPRINT(fprintf(stderr, "Type: ARRAY LBRACK INT_NUM RANGE INT_NUM RBRACK OF StandardType\n"));
    rdtoken->advance(ARRAY);
    synth1=rdoldval;
    rdtoken->advance(LBRACK);
    synth2=rdoldval;
    rdtoken->advance(INT_NUM);
    synth3.Integer=rdoldval.Integer;
    rdtoken->advance(RANGE);
    synth4=rdoldval;
    rdtoken->advance(INT_NUM);
    synth5.Integer=rdoldval.Integer;
    rdtoken->advance(RBRACK);
    synth6=rdoldval;
    rdtoken->advance(OF);
    synth7=rdoldval;
    synth8=parseStandardType(&rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4, &synth5, &synth6, &synth7,&synth8);
    /* Set type and dimension. */
    while((CurrentSymbol = temp_stack.Pop()) != NULL) {
      CurrentSymbol->SetDefinitionType(T_ARRAY);
      CurrentSymbol->SetDimension(synth5.Integer);
      variable_stack.Push(CurrentSymbol);
    }
  }
  rdtoken->getFollow(&RdToken::inFollowType,&RdToken::inAugmentType);
  return *retval;
}

rdunion RdParser::parseStandardType(rdunion *inherit8, rdunion *inherit7, rdunion *inherit6, rdunion *inherit5, rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstStandardType,&RdToken::inFollowStandardType,&RdToken::inAugmentStandardType);
  if(rdtoken->inFirst_INTEGER()) {
    DPRINT(fprintf(stderr, "StandardType: INTEGER\n"));
    rdtoken->advance(INTEGER);
    synth1=rdoldval;
  } else if(rdtoken->inFirst_REAL()) {
    DPRINT(fprintf(stderr, "StandardType: REAL\n"));
    rdtoken->advance(REAL);
    synth1=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowStandardType,&RdToken::inAugmentStandardType);
  return *retval;
}

rdunion RdParser::parseSubprogramDeclarations(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstSubprogramDeclarations,&RdToken::inFollowSubprogramDeclarations,&RdToken::inAugmentSubprogramDeclarations);
  if(rdtoken->inFirst_SubprogramDeclaration_SEMICOLON_SubprogramDeclarations()) {
    DPRINT(fprintf(stderr, "SubprogramDeclarations: SubprogramDeclaration SEMICOLON SubprogramDeclarations\n"));
    synth1=parseSubprogramDeclaration(&rdoldval, &rdoldval,&synth1);
    rdtoken->advance(SEMICOLON);
    synth2=rdoldval;
    /* Insure global variables set, Ensure if over 65.*/
    LL_Node *temp_node;
    global_stack.First();
    while(global_stack.GetCurrent() != NULL) {
      temp_node = global_stack.GetCurrent()->GetSymbol();
      printf(" %s ", temp_node->GetName());
      temp_stack.Push(temp_node);
      global_stack.Next();
    }
    printf(" end\n");
    global_stack.First();
    while((temp_node = temp_stack.Pop()) != NULL) {
      printf(" /%s exch def\n", temp_node->GetName());
    }
    printf("  } store\n");
    synth3=parseSubprogramDeclarations(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "SubprogramDeclarations: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowSubprogramDeclarations,&RdToken::inAugmentSubprogramDeclarations);
  return *retval;
}

rdunion RdParser::parseSubprogramDeclaration(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstSubprogramDeclaration,&RdToken::inFollowSubprogramDeclaration,&RdToken::inAugmentSubprogramDeclaration);
  if(rdtoken->inFirst_SubprogramHead_AllDeclaration_CompoundStatement()) {
    DPRINT(fprintf(stderr, "SubprogramDeclaration: SubprogramHead AllDeclaration CompoundStatement\n"));
    synth1=parseSubprogramHead(&rdoldval, &rdoldval,&synth1);
    synth2=parseAllDeclaration(&rdoldval, &rdoldval, &synth1,&synth2);
    /* No local variables, so we didn't go into AllDeclaration. */
    /* We need to setup up the dictionary for this proc. */
    if(declaration_flag == false) {
      printf(" %i dict begin\n", current_offset);
    }
    /* Do we have parameters??? */
    while((CurrentSymbol = variable_stack.Pop()) != NULL) {
      if(CurrentSymbol->GetDefinitionType() == T_VAR) {
	printf(" /%s exch def \n",CurrentSymbol->GetName());
      } else {
	printf(" %i array astore /%s exch def \n",CurrentSymbol->GetDimension(),CurrentSymbol->GetName());
      }
    }
    synth3=parseCompoundStatement(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    /* Pop the current symbol table. */
    CurrentSymbolTable->FPrintF(stderr);
    CurrentSymbolTable = st_stack.Pop();
  }
  rdtoken->getFollow(&RdToken::inFollowSubprogramDeclaration,&RdToken::inAugmentSubprogramDeclaration);
  return *retval;
}

rdunion RdParser::parseSubprogramHead(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5, synth6;

  rdtoken->getFirst(&RdToken::inFirstSubprogramHead,&RdToken::inFollowSubprogramHead,&RdToken::inAugmentSubprogramHead);
  if(rdtoken->inFirst_FUNCTION_ID_Arguments_COLON_StandardType_SEMICOLON()) {
    DPRINT(fprintf(stderr, "SubprogramHead: FUNCTION ID Arguments COLON StandardType SEMICOLON\n"));
    rdtoken->advance(FUNCTION);
    synth1=rdoldval;
    rdtoken->advance(ID);
    synth2=rdoldval;
    declaration_flag = false;
    current_offset = 0;
    CurrentProc = CurrentSymbolTable->Insert(id_lookback);
    /* Get a new symbol table and push the current onto the stack. */
    CurrentProc->SetDefinitionType(T_FUNCTION);
    strcpy (last_fcn,"");
    strcpy (last_fcn,CurrentProc->GetName());
    CurrentProc->SetBlockLevel(st_stack.GetSize());
    printf(" /%s { ", id_lookback);
    /* Put a new symbol table on the symbol table stack. */
    CurrentSymbolTable = new Symbol_Table;
    st_stack.Push(CurrentSymbolTable);
    synth3=parseArguments(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    CurrentProc->SetSize(current_offset);
    rdtoken->advance(COLON);
    synth4=rdoldval;
    synth5=parseStandardType(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4,&synth5);
    rdtoken->advance(SEMICOLON);
    synth6=rdoldval;
  } else if(rdtoken->inFirst_PROCEDURE_ID_Arguments_SEMICOLON()) {
    DPRINT(fprintf(stderr, "SubprogramHead: PROCEDURE ID Arguments SEMICOLON\n"));
    rdtoken->advance(PROCEDURE);
    synth1=rdoldval;
    rdtoken->advance(ID);
    synth2=rdoldval;
    declaration_flag = false;
    current_offset = 0;
    CurrentProc = CurrentSymbol = CurrentSymbolTable->Insert(id_lookback);
    /* Get a new symbol table and push the current onto the stack. */
    CurrentProc->SetDefinitionType(T_PROCEDURE);
    CurrentProc->SetBlockLevel(st_stack.GetSize());
    printf(" /%s { ", id_lookback);
    /* Put a new symbol table on the symbol table stack. */
    CurrentSymbolTable = new Symbol_Table;
    st_stack.Push(CurrentSymbolTable);
    synth3=parseArguments(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    CurrentProc->SetSize(current_offset);
    rdtoken->advance(SEMICOLON);
    synth4=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowSubprogramHead,&RdToken::inAugmentSubprogramHead);
  return *retval;
}

rdunion RdParser::parseArguments(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstArguments,&RdToken::inFollowArguments,&RdToken::inAugmentArguments);
  if(rdtoken->inFirst_LPAREN_ParameterList_RPAREN()) {
    DPRINT(fprintf(stderr, "Arguments: LPAREN ParameterList RPAREN\n"));
    if(variable_stack.GetSize() > 0) {
      variable_stack.Push(NULL);
    }
    rdtoken->advance(LPAREN);
    synth1=rdoldval;
    synth2=parseParameterList(&rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(RPAREN);
    synth3=rdoldval;
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "Arguments: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowArguments,&RdToken::inAugmentArguments);
  return *retval;
}

rdunion RdParser::parseParameterList(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4;

  rdtoken->getFirst(&RdToken::inFirstParameterList,&RdToken::inFollowParameterList,&RdToken::inAugmentParameterList);
  if(rdtoken->inFirst_IdentifierList_COLON_Type_ParameterTail()) {
    DPRINT(fprintf(stderr, "ParameterList: IdentifierList COLON Type ParameterTail\n"));
    synth1=parseIdentifierList(&rdoldval, &rdoldval, &rdoldval,&synth1);
    rdtoken->advance(COLON);
    synth2=rdoldval;
    synth3=parseType(&rdoldval, &rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    synth4=parseParameterTail(&rdoldval, &rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
  }
  rdtoken->getFollow(&RdToken::inFollowParameterList,&RdToken::inAugmentParameterList);
  return *retval;
}

rdunion RdParser::parseParameterTail(rdunion *inherit5, rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5;

  rdtoken->getFirst(&RdToken::inFirstParameterTail,&RdToken::inFollowParameterTail,&RdToken::inAugmentParameterTail);
  if(rdtoken->inFirst_SEMICOLON_IdentifierList_COLON_Type_ParameterTail()) {
    DPRINT(fprintf(stderr, "ParameterTail: SEMICOLON IdentifierList COLON Type ParameterTail\n"));
    rdtoken->advance(SEMICOLON);
    synth1=rdoldval;
    synth2=parseIdentifierList(&rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(COLON);
    synth3=rdoldval;
    synth4=parseType(&rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
    synth5=parseParameterTail(&rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4,&synth5);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "ParameterTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowParameterTail,&RdToken::inAugmentParameterTail);
  return *retval;
}

rdunion RdParser::parseCompoundStatement(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstCompoundStatement,&RdToken::inFollowCompoundStatement,&RdToken::inAugmentCompoundStatement);
  if(rdtoken->inFirst_BEGIN95BLOCK_OptionalStatement_END95BLOCK()) {
    DPRINT(fprintf(stderr, "CompoundStatement: BEGIN_BLOCK OptionalStatement END_BLOCK\n"));
    rdtoken->advance(BEGIN_BLOCK);
    synth1=rdoldval;
    synth2=parseOptionalStatement(&rdoldval, &rdoldval, &synth1,&synth2);
    strcpy (last_fcn,"empty_");
    rdtoken->advance(END_BLOCK);
    synth3=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowCompoundStatement,&RdToken::inAugmentCompoundStatement);
  return *retval;
}

rdunion RdParser::parseOptionalStatement(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstOptionalStatement,&RdToken::inFollowOptionalStatement,&RdToken::inAugmentOptionalStatement);
  if(rdtoken->inFirst_StatementList()) {
    DPRINT(fprintf(stderr, "OptionalStatement: StatementList\n"));
    synth1=parseStatementList(&rdoldval, &rdoldval,&synth1);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "OptionalStatement: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowOptionalStatement,&RdToken::inAugmentOptionalStatement);
  return *retval;
}

rdunion RdParser::parseStatementList(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstStatementList,&RdToken::inFollowStatementList,&RdToken::inAugmentStatementList);
  if(rdtoken->inFirst_Statement_StatementListTail()) {
    DPRINT(fprintf(stderr, "StatementList: RD_EMPTY\n"));
    synth1=parseStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseStatementListTail(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowStatementList,&RdToken::inAugmentStatementList);
  return *retval;
}

rdunion RdParser::parseStatementListTail(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstStatementListTail,&RdToken::inFollowStatementListTail,&RdToken::inAugmentStatementListTail);
  if(rdtoken->inFirst_SEMICOLON_Statement_StatementListTail()) {
    DPRINT(fprintf(stderr, "StatementListTail: SEMICOLON Statement StatementListTail\n"));
    rdtoken->advance(SEMICOLON);
    synth1=rdoldval;
    synth2=parseStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    synth3=parseStatementListTail(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "StatementListTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowStatementListTail,&RdToken::inAugmentStatementListTail);
  return *retval;
}

rdunion RdParser::parseStatement(rdunion *inherit7, rdunion *inherit6, rdunion *inherit5, rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstStatement,&RdToken::inFollowStatement,&RdToken::inAugmentStatement);
  if(rdtoken->inFirst_ID_StatementOp()) {
    DPRINT(fprintf(stderr, "Statement: ID StatementOp\n"));
    rdtoken->advance(ID);
    synth1=rdoldval;
    CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
    if((CurrentSymbol == NULL)&&(CurrentSymbolTable != GlobalSymbolTable))
      CurrentSymbol = GlobalSymbolTable->Lookup(id_lookback, -1);
    synth2=parseStatementOp(&rdoldval, &rdoldval, &synth1,&synth2);
  } else if(rdtoken->inFirst_IfStatement()) {
    DPRINT(fprintf(stderr, "Statement: IfStatement\n"));
    synth1=parseIfStatement(&rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_CompoundStatement()) {
    DPRINT(fprintf(stderr, "Statement: CompoundStatement\n"));
    synth1=parseCompoundStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_WhileStatement()) {
    DPRINT(fprintf(stderr, "Statement: WhileStatement\n"));
    synth1=parseWhileStatement(&rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_ForStatement()) {
    DPRINT(fprintf(stderr, "Statement: ForStatement\n"));
    synth1=parseForStatement(&rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_ReadStatement()) {
    DPRINT(fprintf(stderr, "Statement: ReadStatement\n"));
    synth1=parseReadStatement(&rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_WriteStatement()) {
    DPRINT(fprintf(stderr, "Statement: WriteStatement\n"));
    synth1=parseWriteStatement(&rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_WriteLnStatement()) {
    DPRINT(fprintf(stderr, "Statement: WriteLnStatement\n"));
    synth1=parseWriteLnStatement(&rdoldval, &rdoldval,&synth1);
  }
  rdtoken->getFollow(&RdToken::inFollowStatement,&RdToken::inAugmentStatement);
  return *retval;
}

rdunion RdParser::parseStatementOp(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstStatementOp,&RdToken::inFollowStatementOp,&RdToken::inAugmentStatementOp);
  if(rdtoken->inFirst_Index_OptionalAssign()) {
    DPRINT(fprintf(stderr, "StatementOp: Index OptionalAssign\n"));
    if(CurrentSymbol == NULL) {
      fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
      compile_errors++;
    } else {
      printf(" %s ", CurrentSymbol->GetName());
    }
    synth1=parseIndex(&rdoldval, &rdoldval,&synth1);
    synth2=parseOptionalAssign(&rdoldval, &rdoldval, &synth1,&synth2);
    printf(" put\n");
  } else if(rdtoken->inFirst_ProcTail_OptionalAssign()) {
    DPRINT(fprintf(stderr, "StatementOp: ProcTail OptionalAssign\n"));
    // :o) storing a procedure name which is not defined
    //procedure_call_stack.Push(CurrentSymbol);
    CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
    if (CurrentSymbol!=NULL) strcpy (id_lookback,"");
    procedure_call_stack.Push(CurrentSymbol);
    strcpy (temp_proc_name,"");
    strcpy (temp_proc_name,id_lookback);
    synth1=parseProcTail(&rdoldval, &rdoldval,&synth1);
    if (CurrentSymbol==NULL)
      {
	if (strcmp(temp_proc_name,last_fcn)!=0)
	  printf("%s \n", temp_proc_name);
      }
    synth2=parseOptionalAssign(&rdoldval, &rdoldval, &synth1,&synth2);
  } else if(rdtoken->inFirst_AssignStatement()) {
    DPRINT(fprintf(stderr, "StatementOp: AssignStatement\n"));
    assign_flag = false;
    if(CurrentSymbol == NULL) {
      fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
      compile_errors++;
    } else {
      /* Check for function return values. */
      if(CurrentSymbol->GetDefinitionType() == T_PROCEDURE) {
	fprintf(stderr, "%i: Attempt to return from procedure: %s\n",rdlines,CurrentSymbol->GetSpelling());
	compile_errors++;
      }
      if(CurrentSymbol->GetDefinitionType() != T_FUNCTION) {
	printf(" /%s ", CurrentSymbol->GetName());
	assign_flag = true;
      }
    }
    synth1=parseAssignStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
    if(assign_flag == true) {
      printf(" store\n");
    }
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "StatementOp: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowStatementOp,&RdToken::inAugmentStatementOp);
  return *retval;
}

rdunion RdParser::parseOptionalAssign(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstOptionalAssign,&RdToken::inFollowOptionalAssign,&RdToken::inAugmentOptionalAssign);
  if(rdtoken->inFirst_AssignStatement()) {
    DPRINT(fprintf(stderr, "OptionalAssign: AssignStatement\n"));
    /* The RHS is the return value from this function. */
    synth1=parseAssignStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "OptionalAssign: RD_EMPTY\n"));
    CurrentSymbol = procedure_call_stack.Pop();
    if(CurrentSymbol != NULL) {
      printf(" %s ",CurrentSymbol->GetSpelling());
    }
  }
  rdtoken->getFollow(&RdToken::inFollowOptionalAssign,&RdToken::inAugmentOptionalAssign);
  return *retval;
}

rdunion RdParser::parseAssignStatement(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstAssignStatement,&RdToken::inFollowAssignStatement,&RdToken::inAugmentAssignStatement);
  if(rdtoken->inFirst_ASSIGN_Expression()) {
    DPRINT(fprintf(stderr, "AssignStatement: ASSIGN Expression\n"));
    rdtoken->advance(ASSIGN);
    synth1=rdoldval;
    synth2=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowAssignStatement,&RdToken::inAugmentAssignStatement);
  return *retval;
}

rdunion RdParser::parseWriteStatement(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4;

  rdtoken->getFirst(&RdToken::inFirstWriteStatement,&RdToken::inFollowWriteStatement,&RdToken::inAugmentWriteStatement);
  if(rdtoken->inFirst_WRITE_LPAREN_Expression_RPAREN()) {
    DPRINT(fprintf(stderr, "WriteStatement: WRITE LPAREN Expression RPAREN\n"));
    rdtoken->advance(WRITE);
    synth1=rdoldval;
    rdtoken->advance(LPAREN);
    synth2=rdoldval;
    synth3=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    rdtoken->advance(RPAREN);
    synth4=rdoldval;
    /* Ok we want to write the top of the stack. */
    printf("write_line_string cvs ");
    printf(" show\n");
  }
  rdtoken->getFollow(&RdToken::inFollowWriteStatement,&RdToken::inAugmentWriteStatement);
  return *retval;
}

rdunion RdParser::parseWriteLnStatement(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4;

  rdtoken->getFirst(&RdToken::inFirstWriteLnStatement,&RdToken::inFollowWriteLnStatement,&RdToken::inAugmentWriteLnStatement);
  if(rdtoken->inFirst_WRITELN_LPAREN_Expression_RPAREN()) {
    DPRINT(fprintf(stderr, "WriteStatement: WRITELN LPAREN Expression RPAREN\n"));
    rdtoken->advance(WRITELN);
    synth1=rdoldval;
    rdtoken->advance(LPAREN);
    synth2=rdoldval;
    synth3=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    rdtoken->advance(RPAREN);
    synth4=rdoldval;
    /* Ok we want to write the top of the stack. */
    printf("write_line_string cvs ");
    printf(" show\n");
    printf(" /CP CP fontsize sub 1 sub def\n");
    printf(" LM CP moveto\n");
  }
  rdtoken->getFollow(&RdToken::inFollowWriteLnStatement,&RdToken::inAugmentWriteLnStatement);
  return *retval;
}

rdunion RdParser::parseReadStatement(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4;

  rdtoken->getFirst(&RdToken::inFirstReadStatement,&RdToken::inFollowReadStatement,&RdToken::inAugmentReadStatement);
  if(rdtoken->inFirst_READ_LPAREN_IdRList_RPAREN()) {
    DPRINT(fprintf(stderr, "ReadStatement: READ LPAREN IdRList RPAREN\n"));
    rdtoken->advance(READ);
    synth1=rdoldval;
    rdtoken->advance(LPAREN);
    synth2=rdoldval;
    synth3=parseIdRList(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    rdtoken->advance(RPAREN);
    synth4=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowReadStatement,&RdToken::inAugmentReadStatement);
  return *retval;
}

rdunion RdParser::parseIdRList(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstIdRList,&RdToken::inFollowIdRList,&RdToken::inAugmentIdRList);
  if(rdtoken->inFirst_IdAccess_IdRTail()) {
    DPRINT(fprintf(stderr, "IdRList: IdAccess IdRTail\n"));
    synth1=parseIdAccess(&rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseIdRTail(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowIdRList,&RdToken::inAugmentIdRList);
  return *retval;
}

rdunion RdParser::parseIdRTail(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstIdRTail,&RdToken::inFollowIdRTail,&RdToken::inAugmentIdRTail);
  if(rdtoken->inFirst_COMMA_IdAccess_IdRTail()) {
    DPRINT(fprintf(stderr, "IdRTail: COMMA IdAccess IdRTail\n"));
    rdtoken->advance(COMMA);
    synth1=rdoldval;
    synth2=parseIdAccess(&rdoldval, &rdoldval, &synth1,&synth2);
    synth3=parseIdRTail(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "IdRTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowIdRTail,&RdToken::inAugmentIdRTail);
  return *retval;
}

rdunion RdParser::parseIndex(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstIndex,&RdToken::inFollowIndex,&RdToken::inAugmentIndex);
  if(rdtoken->inFirst_LBRACK_IndexType_RBRACK()) {
    DPRINT(fprintf(stderr, "Index: LBRACK IndexType RBRACK\n"));
    rdtoken->advance(LBRACK);
    synth1=rdoldval;
    synth2=parseIndexType(&rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(RBRACK);
    synth3=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowIndex,&RdToken::inAugmentIndex);
  return *retval;
}

rdunion RdParser::parseIndexType(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstIndexType,&RdToken::inFollowIndexType,&RdToken::inAugmentIndexType);
  if(rdtoken->inFirst_INT95NUM()) {
    DPRINT(fprintf(stderr, "IndexType: INT_NUM\n"));
    rdtoken->advance(INT_NUM);
    synth1.Integer=rdoldval.Integer;
    printf(" %i 1 sub\n", synth1.Integer);
  } else if(rdtoken->inFirst_ID()) {
    DPRINT(fprintf(stderr, "IndexType: ID\n"));
    rdtoken->advance(ID);
    synth1=rdoldval;
    CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
    if((CurrentSymbol == NULL)&&(CurrentSymbolTable != GlobalSymbolTable))
      CurrentSymbol = GlobalSymbolTable->Lookup(id_lookback, -1);
    if(CurrentSymbol == NULL) {
      fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
      compile_errors++;
    } else {
      printf(" %s 1 sub\n", CurrentSymbol->GetName());
    }
  }
  rdtoken->getFollow(&RdToken::inFollowIndexType,&RdToken::inAugmentIndexType);
  return *retval;
}

rdunion RdParser::parseIdAccess(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstIdAccess,&RdToken::inFollowIdAccess,&RdToken::inAugmentIdAccess);
  if(rdtoken->inFirst_ID_IdAccessTail()) {
    DPRINT(fprintf(stderr, "IdAccess: ID IdAccessTail\n"));
    rdtoken->advance(ID);
    synth1=rdoldval;
    synth2=parseIdAccessTail(&rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowIdAccess,&RdToken::inAugmentIdAccess);
  return *retval;
}

rdunion RdParser::parseIdAccessTail(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstIdAccessTail,&RdToken::inFollowIdAccessTail,&RdToken::inAugmentIdAccessTail);
  if(rdtoken->inFirst_Index()) {
    DPRINT(fprintf(stderr, "IdAccessTail: Index\n"));
    synth1=parseIndex(&rdoldval, &rdoldval,&synth1);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "IdAccessTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowIdAccessTail,&RdToken::inAugmentIdAccessTail);
  return *retval;
}

rdunion RdParser::parseWhileStatement(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4;

  rdtoken->getFirst(&RdToken::inFirstWhileStatement,&RdToken::inFollowWhileStatement,&RdToken::inAugmentWhileStatement);
  if(rdtoken->inFirst_WHILE_Expression_DO_Statement()) {
    printf("{ %% beginning of while loop\n");
    DPRINT(fprintf(stderr, "WhileStatement: WHILE Expression DO \n"));
    rdtoken->advance(WHILE);
    synth1=rdoldval;
    synth2=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(DO);
    synth3=rdoldval;
    printf("\n  { ");
    DPRINT(fprintf(stderr, "WhileStatement: Statement\n"));
    synth4=parseStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
    printf("  }\n {exit} ifelse\n} loop %%  end of while\n");
  }
  rdtoken->getFollow(&RdToken::inFollowWhileStatement,&RdToken::inAugmentWhileStatement);
  return *retval;
}

rdunion RdParser::parseForStatement(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5, synth6, synth7;

  rdtoken->getFirst(&RdToken::inFirstForStatement,&RdToken::inFollowForStatement,&RdToken::inAugmentForStatement);
  if(rdtoken->inFirst_FOR_ID_AssignStatement_TO_Expression_DO_Statement()) {
    rdtoken->advance(FOR);
    synth1=rdoldval;
    rdtoken->advance(ID);
    synth2=rdoldval;
    synth3=parseAssignStatement(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    rdtoken->advance(TO);
    synth4=rdoldval;
    printf("1 "); //assuming step is 1
    synth5=parseExpression(&rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4,&synth5);
    rdtoken->advance(DO);
    synth6=rdoldval;
    printf("{");
    synth7=parseStatement(&rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4, &synth5, &synth6,&synth7);
    printf("} for \n");
  }
  rdtoken->getFollow(&RdToken::inFollowForStatement,&RdToken::inAugmentForStatement);
  return *retval;
}

rdunion RdParser::parseIfStatement(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4, synth5;

  rdtoken->getFirst(&RdToken::inFirstIfStatement,&RdToken::inFollowIfStatement,&RdToken::inAugmentIfStatement);
  if(rdtoken->inFirst_IF_Expression_THEN_Statement_Testelse()) {
    DPRINT(fprintf(stderr, "IfStatement: IF Expression THEN \n"));
    rdtoken->advance(IF);
    synth1=rdoldval;
    synth2=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(THEN);
    synth3=rdoldval;
    printf("{");
    synth4=parseStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
    printf("}\n");
    synth5=parseTestelse(&rdoldval, &rdoldval, &synth1, &synth2, &synth3, &synth4,&synth5);
  }
  rdtoken->getFollow(&RdToken::inFollowIfStatement,&RdToken::inAugmentIfStatement);
  return *retval;
}

rdunion RdParser::parseTestelse(rdunion *inherit5, rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstTestelse,&RdToken::inFollowTestelse,&RdToken::inAugmentTestelse);
  if(rdtoken->inFirst_ELSE_Statement()) {
    DPRINT(fprintf(stderr, "else or no else \n"));
    rdtoken->advance(ELSE);
    synth1=rdoldval;
    printf("{");
    synth2=parseStatement(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    printf("}\n");
    printf(" ifelse ");
  } else if(rdtoken->inFirst_SEMICOLON()) {
    rdtoken->advance(SEMICOLON);
    synth1=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowTestelse,&RdToken::inAugmentTestelse);
  return *retval;
}

rdunion RdParser::parseProcTail(rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstProcTail,&RdToken::inFollowProcTail,&RdToken::inAugmentProcTail);
  if(rdtoken->inFirst_LPAREN_ProcExpressionList_RPAREN()) {
    DPRINT(fprintf(stderr, "ProcTail: LPAREN ProcExpressionList RPAREN\n"));
    rdtoken->advance(LPAREN);
    synth1=rdoldval;
    synth2=parseProcExpressionList(&rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(RPAREN);
    synth3=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowProcTail,&RdToken::inAugmentProcTail);
  return *retval;
}

rdunion RdParser::parseProcExpressionList(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstProcExpressionList,&RdToken::inFollowProcExpressionList,&RdToken::inAugmentProcExpressionList);
  if(rdtoken->inFirst_Expression_ProcExpressionListTail()) {
    DPRINT(fprintf(stderr, "ProcExpressionList: Expression ProcExpressionListTail\n"));
    synth1=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseProcExpressionListTail(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    while((CurrentSymbol = function_stack.Pop()) != NULL) {
      printf(" %s ",CurrentSymbol->GetName());
    }
    while((CurrentSymbol = procedure_call_stack.Pop()) != NULL){
      printf(" %s ",CurrentSymbol->GetName());
    }
  }
  rdtoken->getFollow(&RdToken::inFollowProcExpressionList,&RdToken::inAugmentProcExpressionList);
  return *retval;
}

rdunion RdParser::parseProcExpressionListTail(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstProcExpressionListTail,&RdToken::inFollowProcExpressionListTail,&RdToken::inAugmentProcExpressionListTail);
  if(rdtoken->inFirst_COMMA_Expression_ProcExpressionListTail()) {
    DPRINT(fprintf(stderr, "ProcExpressionListTail: COMMA Expression ProcExpressionListTail\n"));
    rdtoken->advance(COMMA);
    synth1=rdoldval;
    synth2=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    synth3=parseProcExpressionListTail(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "ProcExpressionListTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowProcExpressionListTail,&RdToken::inAugmentProcExpressionListTail);
  return *retval;
}

rdunion RdParser::parseExpression(rdunion *inherit5, rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstExpression,&RdToken::inFollowExpression,&RdToken::inAugmentExpression);
  if(rdtoken->inFirst_SimpleExpression_ExpressionTail()) {
    DPRINT(fprintf(stderr, "Expression: SimpleExpression ExpressionTail\n"));
    synth1=parseSimpleExpression(&rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseExpressionTail(&rdoldval, &rdoldval, &synth1,&synth2);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "Expression: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowExpression,&RdToken::inAugmentExpression);
  return *retval;
}

rdunion RdParser::parseExpressionTail(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstExpressionTail,&RdToken::inFollowExpressionTail,&RdToken::inAugmentExpressionTail);
  if(rdtoken->inFirst_RELOP_SimpleExpression()) {
    DPRINT(fprintf(stderr, "ExpressionTail: RELOP SimpleExpression\n"));
    rdtoken->advance(RELOP);
    synth1.Op=rdoldval.Op;
    synth2=parseSimpleExpression(&rdoldval, &rdoldval, &synth1,&synth2);
    switch(synth1.Op) {
    case EQ:
      printf(" eq ");
      break;
    case GE:
      printf(" ge ");
      break;
    case GT:
      printf(" gt ");
      break;
    case LT:
      printf(" lt ");
      break;
    case LE:
      printf(" le ");
      break;
    case NE:
      printf(" ne ");
      break;
    }
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "ExpressionTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowExpressionTail,&RdToken::inAugmentExpressionTail);
  return *retval;
}

rdunion RdParser::parseSimpleExpression(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstSimpleExpression,&RdToken::inFollowSimpleExpression,&RdToken::inAugmentSimpleExpression);
  if(rdtoken->inFirst_SIGN_Term_SimpleExpressionTail()) {
    DPRINT(fprintf(stderr, "SimpleExpression: SIGN Term SimpleExpressionTail\n"));
    rdtoken->advance(SIGN);
    synth1.Op=rdoldval.Op;
    synth2=parseTerm(&rdoldval, &rdoldval, &synth1,&synth2);
    synth3=parseSimpleExpressionTail(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    if(synth1.Op == MINUS) {
      printf(" neg\n");
    }
  } else if(rdtoken->inFirst_Term_SimpleExpressionTail()) {
    DPRINT(fprintf(stderr, "SimpleExpression: Term SimpleExpressionTail\n"));
    synth1=parseTerm(&rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseSimpleExpressionTail(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowSimpleExpression,&RdToken::inAugmentSimpleExpression);
  return *retval;
}

rdunion RdParser::parseSimpleExpressionTail(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstSimpleExpressionTail,&RdToken::inFollowSimpleExpressionTail,&RdToken::inAugmentSimpleExpressionTail);
  if(rdtoken->inFirst_ADDOP_Term_SimpleExpressionTail()) {
    DPRINT(fprintf(stderr, "SimpleExpressionTail: ADDOP Term SimpleExpressionTail\n"));
    rdtoken->advance(ADDOP);
    synth1.Op=rdoldval.Op;
    synth2=parseTerm(&rdoldval, &rdoldval, &synth1,&synth2);
    synth3=parseSimpleExpressionTail(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    switch(synth1.Op) {
    case PLUS:
      printf(" add\n");
      break;
    case MINUS:
      printf(" sub\n");
      break;
    case OR:
      printf(" or\n");
      break;
    }
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "SimpleExpressionTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowSimpleExpressionTail,&RdToken::inAugmentSimpleExpressionTail);
  return *retval;
}

rdunion RdParser::parseTerm(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2;

  rdtoken->getFirst(&RdToken::inFirstTerm,&RdToken::inFollowTerm,&RdToken::inAugmentTerm);
  if(rdtoken->inFirst_Factor_TermTail()) {
    DPRINT(fprintf(stderr, "Term: Factor TermTail\n"));
    synth1=parseFactor(&rdoldval, &rdoldval, &rdoldval, &rdoldval,&synth1);
    synth2=parseTermTail(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
  }
  rdtoken->getFollow(&RdToken::inFollowTerm,&RdToken::inAugmentTerm);
  return *retval;
}

rdunion RdParser::parseSign(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstSign,&RdToken::inFollowSign,&RdToken::inAugmentSign);
  if(rdtoken->inFirst_SIGN()) {
    DPRINT(fprintf(stderr, "Sign: SIGN\n"));
    rdtoken->advance(SIGN);
    synth1.Op=rdoldval.Op;
    if(synth1.Op == MINUS) {
      printf(" neg\n");
    }
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "Sign: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowSign,&RdToken::inAugmentSign);
  return *retval;
}

rdunion RdParser::parseTermTail(rdunion *inherit4, rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3, synth4;

  rdtoken->getFirst(&RdToken::inFirstTermTail,&RdToken::inFollowTermTail,&RdToken::inAugmentTermTail);
  if(rdtoken->inFirst_MULOP_Sign_Factor_TermTail()) {
    DPRINT(fprintf(stderr, "TermTail: MULOP Sign Factor TermTail\n"));
    rdtoken->advance(MULOP);
    synth1.Op=rdoldval.Op;
    synth2=parseSign(&rdoldval, &rdoldval, &synth1,&synth2);
    synth3=parseFactor(&rdoldval, &rdoldval, &synth1, &synth2,&synth3);
    synth4=parseTermTail(&rdoldval, &rdoldval, &synth1, &synth2, &synth3,&synth4);
    switch(synth1.Op) {
    case MULTIPLY:
      printf(" mul\n");
      break;
    case DIVIDE:
      printf(" div\n");
      break;
    case MOD:
      printf(" mod\n");
      break;
    case DIV:
      printf(" idiv\n");
      break;
    case AND:
      printf(" and\n");
      break;
    }
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "TermTail: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowTermTail,&RdToken::inAugmentTermTail);
  return *retval;
}

rdunion RdParser::parseFactor(rdunion *inherit3, rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1, synth2, synth3;

  rdtoken->getFirst(&RdToken::inFirstFactor,&RdToken::inFollowFactor,&RdToken::inAugmentFactor);
  if(rdtoken->inFirst_ID_FactorOp()) {
    DPRINT(fprintf(stderr, "Factor: ID FactorOp\n"));
    rdindex = false;
    CurrentSymbol = CurrentSymbolTable->Lookup(id_lookback, -1);
    if((CurrentSymbol == NULL)&&(CurrentSymbolTable != GlobalSymbolTable))
      CurrentSymbol = GlobalSymbolTable->Lookup(id_lookback, -1);
    if(CurrentSymbol == NULL) {
      fprintf(stderr, "%i: Undefined identifier: %s\n",rdlines,id_lookback);
      compile_errors++;
    } else {
      if(CurrentSymbol->GetDefinitionType() != T_FUNCTION && CurrentSymbol->GetDefinitionType() != T_PROCEDURE) {
	printf(" %s ",CurrentSymbol->GetName());
      }
      // pushing function names into a function stack
      if (CurrentSymbol->GetDefinitionType() ==T_FUNCTION)
	{
	  function_stack.Push(CurrentSymbol);
	}
      // pushing procedure names into a procedure stack
      if (CurrentSymbol->GetDefinitionType() ==T_PROCEDURE)
	{
	  procedure_call_stack.Push(CurrentSymbol);
	}
    }
    rdtoken->advance(ID);
    synth1=rdoldval;
    synth2=parseFactorOp(&rdoldval, &rdoldval, &synth1,&synth2);
    if(rdindex == true) {
      /* I think the value of index should be on the PS stack. */
      printf(" get ");
    }
  } else if(rdtoken->inFirst_NOT_Factor()) {
    DPRINT(fprintf(stderr, "Factor: NOT Factor\n"));
    rdtoken->advance(NOT);
    synth1=rdoldval;
    synth2=parseFactor(&rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    printf(" not\n");
  } else if(rdtoken->inFirst_INT95NUM()) {
    DPRINT(fprintf(stderr, "Factor: INT_NUM\n"));
    rdtoken->advance(INT_NUM);
    synth1.Integer=rdoldval.Integer;
    printf(" %i ",synth1.Integer);
  } else if(rdtoken->inFirst_REAL95NUM()) {
    DPRINT(fprintf(stderr, "Factor: REAL_NUM\n"));
    rdtoken->advance(REAL_NUM);
    synth1=rdoldval;
    printf(" %g ",synth1.Float);
  } else if(rdtoken->inFirst_LPAREN_Expression_RPAREN()) {
    DPRINT(fprintf(stderr, "Factor: LPAREN Expression RPAREN\n"));
    rdtoken->advance(LPAREN);
    synth1=rdoldval;
    synth2=parseExpression(&rdoldval, &rdoldval, &rdoldval, &rdoldval, &rdoldval, &synth1,&synth2);
    rdtoken->advance(RPAREN);
    synth3=rdoldval;
  }
  rdtoken->getFollow(&RdToken::inFollowFactor,&RdToken::inAugmentFactor);
  return *retval;
}

rdunion RdParser::parseFactorOp(rdunion *inherit2, rdunion *inherit1, rdunion *inherit0, rdunion *retval) {
  rdunion synth1;

  rdtoken->getFirst(&RdToken::inFirstFactorOp,&RdToken::inFollowFactorOp,&RdToken::inAugmentFactorOp);
  if(rdtoken->inFirst_Index()) {
    DPRINT(fprintf(stderr, "FactorOp: Index\n"));
    rdindex = true;
    synth1=parseIndex(&rdoldval, &rdoldval,&synth1);
  } else if(rdtoken->inFirst_ProcTail()) {
    DPRINT(fprintf(stderr, "FactorOp: ProcTail\n"));
    synth1=parseProcTail(&rdoldval, &rdoldval,&synth1);
  } else {
    // | Empty
    DPRINT(fprintf(stderr, "FactorOp: RD_EMPTY\n"));
  }
  rdtoken->getFollow(&RdToken::inFollowFactorOp,&RdToken::inAugmentFactorOp);
  return *retval;
}

