/****************************************************************************
* FILE : sym_tab.cc
* INPUT/OUTPUT  : rd++ input file (partial)
* PROJECT: CS 541 CS-University of Kentucky
* SOURCE:   Provided as a handout by Jerzy W. Jaromczyk (based on projects)
* COPYRIGHT: University of Kentucky @1999
****************************************************************************/

/* fixed for modern compilers by Dirk Jagdmann <doj@cubic.org> */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <iostream>
#include "sym_table.h"
#include "hash.h"
#include "rddefs.h"

extern Symbol_Table *CurrentSymbolTable;
extern char *DEFINED_TYPES[];
extern int scope_level;

/* LL_Node constructor initializes pointers for the LL_Node class*/
LL_Node::LL_Node()
{
	spelling_start = name = NULL;
	next = NULL;
	definition_type = 0;
	size = dimension = 0;
	base_name = NULL;
	symbol_list = NULL;
	offset = 0;  // Offset of current node variable.
}

LL_Node::~LL_Node()
{
  free(name);
}


void LL_Node::FPrintF()
{
  LL_Node *list;

  switch(GetDefinitionType()) {
     case T_VAR:
       printf("  T_VAR:\n");
       list = (LL_Node *)GetSymbolList();
       if(list != NULL) list->FPrintF();
       break;
     case T_ARRAY:
       printf("  T_ARRAY:\n");
       printf("    GetDimension(): %i\n",GetDimension());
       list = (LL_Node *)GetSymbolList();
       if(list != NULL) list->FPrintF();
       break;
     case T_RECORD:
       printf("  T_RECORD:\n");
       list = (LL_Node *)GetSymbolList();
       if(list != NULL) {
	 printf("     GetSymbolList() != NULL)\n");
       } else {
	 printf("     GetSymbolList() == NULL)\n");
       }
       break;
     default:
       printf("  T_UNKNOWN:\n");
       printf("    GetSpelling(): %s\n",GetSpelling());
       list = (LL_Node *)GetSymbolList();
       if(list != NULL) list->FPrintF();
       break;
  }
}

/* Set data in new LL_Nodes */
void LL_Node::SetSpelling(char *start)
{
	spelling_start = start;
}

char *LL_Node::GetName()
{
  /* 40 should be enough for 2 ints. */
  char temp[strlen(spelling_start) + 40];

  /* GetSpelling should be used for funcs and procs, but just incase. */
  if(definition_type == T_FUNCTION || definition_type == T_PROCEDURE) {
    return spelling_start;
  }

  if(name == NULL) {
    sprintf(temp, "%s_%i_%i",spelling_start,block_level,offset);
    if((name = (char *)malloc(strlen(temp) + 1)) == NULL) {
       perror("LL_Node::GetName()");
    }
    memcpy(name, temp, strlen(temp) + 1);
  }

  return name;
}

/* L_List constructor initializes pointers and size. */
L_List::L_List()
{
    last = current_pos = head = NULL;
    pos_num = size = 0;
}

/* L_List destuctor deletes pointers to nodes */
L_List::~L_List()
{
  //Clear();
}

/* Symbol_Table Constructor. */
Symbol_Table::Symbol_Table()
{
       table_usage = 0;
       lastchar = Spelling_Table;
       lookup_start = NULL;
       declaration = FALSE;
       globldeclaration = FALSE;
       globalbegin = FALSE;
}
/* Symbol_Table Destructor. */
Symbol_Table::~Symbol_Table()
{
  //printf("Symbol_Table Destroyed.\n");
}

/* Link list deletion function, frees memory. */
void L_List::Clear()
{
	LL_Node *prev;

	while(head != NULL){
		prev = head;
		head = head->Next();
		delete prev;
	}
	current_pos = head = NULL;
	pos_num = size = 0;
}

/* Make current_pos the last node in the list */
void L_List::First()
{
  current_pos = head;
  pos_num = 0;
}

/* Make current_pos the last node in the list */
void L_List::Last()
{
  current_pos = last;
  if(last != NULL) {
    pos_num = (size - 1);
  } else pos_num = 0;
}

/* Make current_pos the node following the current node. */
void L_List::Next()
{
   if(current_pos != NULL) {
	current_pos = current_pos->Next();
	pos_num++;
   }
}

/* Insert a symbol at the head of a list.*/
void L_List::Insert(char *start)
{
   LL_Node *new_node;

   new_node = new LL_Node;
   new_node->SetSpelling(start);
   if(head == NULL) {
     new_node->SetNext(head);
     current_pos = last = head = new_node;
     pos_num = 0;
     size = 1;
   } else {
     new_node->SetNext(last->Next());
     last->SetNext(new_node);
     last = current_pos = new_node;
     pos_num = size;
     size++;
   }

}

/* Search a list for a symbol. */
LL_Node *L_List::Lookup(char *token)
{
   LL_Node *p;
   char *start;

   p = head;
   while(p != NULL) {
     start = p->GetSpelling();
     if(strcmp(start, token) == 0) {
       current_pos = p;
       return p;
     }
     p = p->Next();
   }
   return NULL;

}

/*  Set the offset for the current token  */
void LL_Node::SetOffset(int new_offset) {
  offset = new_offset;
}


/* Insert a symbol into the table. */
LL_Node *Symbol_Table::Insert(char *new_token)
{
  char *start;
  int token_length, hash_value;
  LL_Node *Symbol;

  hash_value = hash(new_token);
  if(Lookup(new_token, hash_value) != NULL)  {
  	if ((globldeclaration == TRUE) && ((lookahead == BEGIN_BLOCK) || (lookahead==PROCEDURE)))
	        globldeclaration = FALSE;

	// If the current line is the declarations line and token is already
	// in the symbol table, then indicate this is a duplicate token
   	if ((globldeclaration == TRUE ) && (lookahead != COLON)) {
		printf("Error:%i: redeclaration of %s\n",rdlines, new_token);
		compile_errors++;
	}

 	if (lookahead==BEGIN_BLOCK)
	        globalbegin = TRUE;

       // If the current line of code contains the END statement, reset globalbegin
       if ((lookahead == END_BLOCK) || (lookahead== PROCEDURE))
		globalbegin = FALSE;

       return SymbolTable[hash_value].GetCurrent();
  } else {
        // If this token is inside the body of the Procedure/Module, and the ident.
        // was not found in the symbol table, then it has not been defined

        // If this is the end of the declaration then turn globaldeclaration OFF

        if ((globldeclaration == TRUE) && ((lookahead == BEGIN_BLOCK) || (lookahead==PROCEDURE)))
	         globldeclaration = FALSE;

        if (lookahead==BEGIN_BLOCK)
		 globalbegin = TRUE;

        // If the current line of code contains the END statement, reset globalbegin
	if ((lookahead == END_BLOCK) || (lookahead== PROCEDURE))
		 globalbegin = FALSE;

        if ((globalbegin==TRUE) && (globldeclaration!=TRUE) && (lookahead != PROGRAM))
	        if (lookahead!=PROCEDURE) {
		     printf("Error:%i: undefined symbol %s\n",rdlines,new_token);
		     compile_errors++;
		}
  }
  start = lastchar;
  token_length = strlen(new_token) + 1;
  table_usage += token_length;
  if(table_usage >= SPELLING_TABLE_SIZE) {
    fprintf(stderr,"Symbol_Table::Insert(): Spelling Table at Capacity.\n");
    fprintf(stderr,"Symbol_Table::Insert(): Cannot Insert.\n");
    return NULL;
  }
   if (lookahead == VAR)  {
	   declaration = TRUE;
	   globldeclaration = TRUE;
   }
  lastchar += token_length;
  memcpy(start, new_token, token_length);

  SymbolTable[hash_value].Insert(start);
  lastchar++;

  Symbol = SymbolTable[hash_value].GetCurrent();
  Symbol->SetLineNumber(rdlines);

  return Symbol;

}

/* Search for a symbol in the table. */
LL_Node *Symbol_Table::Lookup(char *new_token, int hash_value)
{
  LL_Node *tmp;

  if(hash_value < 0) hash_value = hash(new_token);

  if((tmp = SymbolTable[hash_value].Lookup(new_token)) != NULL) {
    lookup_start = tmp->GetSpelling();
    return tmp;
  } else {
    return NULL;
  }
}

/* Print all the symbols in the table. */
void Symbol_Table::FPrintF(FILE *outstream)
{
  LL_Node *p;
  char *start;
  int i;

  if(outstream == NULL) {
    std::cerr << "Symbol_Table::FPrintF()--> NULL stream parameter." << std::endl;
    return;
  }
  fprintf(outstream, "\n----------Debugging Information.----------\n");
  fprintf(outstream, "[Hash ] [List    ] [List] [Line  ]               [Base]     [Block]\n");
  fprintf(outstream, "[Entry] [Position] [Size] [Number] [Type]        [Name]     [Level] [Offset] Name\n\n");
  for(i = 0; i < SYMBOL_TABLE_SIZE; i++) {
    SymbolTable[i].First();
    p = SymbolTable[i].GetCurrent();
    while(p != NULL) {
      fprintf(outstream,"%-9i ", i);
      fprintf(outstream,"%-9i ",SymbolTable[i].GetPosNum());
      fprintf(outstream,"%-7i ",SymbolTable[i].GetSize());
      fprintf(outstream,"%-7i ",p->GetLineNumber());
      switch(p->GetDefinitionType()) {
          case T_PROCEDURE:
    	       fprintf(outstream,"%-11s","T_PROCEDURE");
 	       break;
          case T_FUNCTION:
    	       fprintf(outstream,"%-11s","T_FUNCTION");
               break;
          case T_VAR:
	       fprintf(outstream,"%-11s","T_VAR");
	       break;
          case T_TYPE:
	       fprintf(outstream,"%-11s","T_TYPE");
	       break;
          case T_ARRAY:
	       fprintf(outstream,"%-11s","T_ARRAY");
	       break;
          case T_RECORD:
	       fprintf(outstream,"%-11s","T_RECORD");
	       break;
          default:
	       fprintf(outstream,"%-11s","UNKNOWN");
	       break;
      }

      if(p->GetBaseName() == NULL) {
	fprintf(outstream,"  %-12s","NULL");
      } else {
	if(p->GetDimension() > 1) {
	  fprintf(outstream,"  %s[%i]%-3s",p->GetBaseName(),p->GetDimension(),"");
	} else {
	  fprintf(outstream,"  %-12s",p->GetBaseName());
	}
      }
      fprintf (outstream, "  %-6i", p->GetBlockLevel());
      fprintf (outstream ,"  %-11i", p->GetOffset());
      start = p->GetSpelling();
      while(*start) {
           fprintf(outstream,"%c",*start);
	   start++;
      }
      fprintf(outstream,"\n");

      //if(p->GetDefinitionType() == T_ARRAY) p->FPrintF();

      SymbolTable[i].Next();
      p = SymbolTable[i].GetCurrent();
    }
  }
  fprintf(outstream,"\n");
}

/* Start an interactive query of the table. */
void Symbol_Table::QueryMode(void)
{
  char token[255];
  int counter = 0, hash_val;

  std::cout << "-------------Query Mode--------------" << std::endl;
  while(1) {
    std::cout << "\nCTRL-C to end Query Mode." << std::endl;
    std::cout << "Enter Token " << counter << "> ";
    std::cin.getline(token, 255, '\n');
    std::cout << "[Hash Entry] [List Position] [List Size] Name" << std::endl;
    hash_val = hash(token);
    if(Lookup(token, hash_val) != NULL) {
      fprintf(stdout,"     %-12i %-10i %-10i %s",
	     hash_val,SymbolTable[hash_val].GetPosNum(),SymbolTable[hash_val].GetSize(),token);
    } else {
      std::cout << token << ": is NOT a valid symbol" << std::endl;
    }
    counter++;
  }
}

